/*********************************************************************

        Sensirion SHT-71用の簡易湿度計 for Arduino
        (SHT-10 SHT-11 SHT-15 SHT-71 SHT-75)

本ソースリストおよびソフトウェアは、ライセンスフリーです。
個人での利用は自由に行えます。著作権表示の改変は禁止します。

                               Copyright (c) 2012 Wataru KUNINO
                               http://www.geocities.jp/bokunimowakaru/
*********************************************************************/
/*
参考文献：
(SENSIRION社用 IF)
・2線式 温・湿度センサSHT11	中尾 司
	http://www.eleki-jack.com/mycom2/pic/cat94/2_sht11/
・データシート：SHT7x（SHT71、SHT75）
	http://www.sensirion.co.jp/pdf/doc_center/01_humidity/01_SHT/jp/Datasheet_SHT7x_V5.0a_C1_J.pdf
・Arduinoをはじめよう (オライリージャパン)
	Massimo Banzi著 船田功訳
・Arduino 日本語リファレンス
	http://www.musashinodenpa.com/arduino/ref/index.php
*/

#include <SD.h>

#define	portSCK			19							// SCKポート番号(ANALOG 5)
#define	portDATA		18							// DATAポート番号(ANALOG 4)
#define	sensorVDD		3.3							// センサー電圧

// SCKを"H"レベルに設定する関数
void SCK_H(void) {
    digitalWrite(portSCK, HIGH);
    delayMicroseconds(50);
}
// SCKを"L"レベルに設定する関数
void SCK_L(void) {
    digitalWrite(portSCK, LOW);
    delayMicroseconds(50);
}

// DATAを"H"レベルに設定する関数
void DATA_H(void) {
    digitalWrite(portDATA, HIGH);
    delayMicroseconds(50);
}
// DATAを"L"レベルに設定する関数
void DATA_L(void) {
    digitalWrite(portDATA, LOW);
    delayMicroseconds(50);
}

// SHTポートの初期化
void SHTInit(void) {
	pinMode(portSCK,OUTPUT);
	pinMode(portDATA,OUTPUT);
}

// SHT TSシーケンス
void SHTTSSeq(void) {
	pinMode(portSCK,OUTPUT);
	pinMode(portDATA,OUTPUT);
    DATA_H();    SCK_H();
    DATA_L();    SCK_L();    SCK_H();
    DATA_H();    SCK_L();
}

// バイト・データ送信
byte SHTWrite(byte dat) {
    byte i, rdat;

	pinMode(portDATA,OUTPUT);
    // DATAビットの出力
    for(i = 0; i < 8; i++) {
        SCK_L();
        if(dat & 0x80) {        // MSB
            // "1"
            DATA_H();
        } else {
            // "0"
            DATA_L();
        }
        SCK_H();
 
        dat <<= 1;              // 次のビット
    }
 
    SCK_L();
 

    // ACKビットの入力
    pinMode(portDATA,INPUT);                   // 入力に切り替え
 
    SCK_H();                    // ↑edge
    if( digitalRead(portDATA) ) {      // read DATA
        // 1
        rdat = 1;               // NOACK
    } else {
        // 0
        rdat = 0;               // ACK
    }
    SCK_L();
 
    return rdat;
}

//バイト・データ受信
byte SHTRead(byte ack) {
    byte i, dat;
 
    dat = 0;
    pinMode(portDATA,INPUT);                  // 入力に切り替え
 
    // DATAビットの入力
    for(i = 0; i < 8; i++) {
        dat <<= 1;                                         // 次のビット
        SCK_H();                                           // ↑edge
        if( digitalRead(portDATA) ) {          // read DATA
            // "1"
            dat |= 1;
        }
        SCK_L();
    }

    
    // ACKビットの出力
    if(!ack) {
        pinMode(portDATA,OUTPUT);               // 出力に切り替え
        DATA_L();
    }
    SCK_H();
    SCK_L();                    // ↓edge

    return dat;
}

float CalcTMP(const unsigned int ADValTmp){
	float d1;
	
	if( 		sensorVDD > 4.5 ){
	//	vdd =   5.0;
		d1  = -40.1;
	}
	else if(	sensorVDD > 3.7 ){
	//	vdd =   4.0;
		d1  = -39.8;
	}
	else if(	sensorVDD > 3.2 ){
	//	vdd =   3.5;
		d1  = -39.7;
	}
	else if(	sensorVDD > 2.7 ){
	//	vdd =   3.0;
		d1  = -39.6;
	}
	else {
	//	vdd =   2.5;
		d1  = -39.4;
	}
	
	return( d1 + 0.01 * (float)ADValTmp );
}

unsigned int CalcTMP10(const unsigned int ADValTmp){
	return (ADValTmp - 4000+5)/10;		// 温度変換(10倍値)
}

float CalcHR(const unsigned int ADValHR , const float temp) {
	float hum = -1.5955e-6 * (float)ADValHR * (float)ADValHR;
	hum += 0.0367 * (float)ADValHR;
	hum -= 2.0468;
	hum += (temp - 25) * (0.01 + 0.00008 * (float)ADValHR );
	return( hum );
}

unsigned int CalcHR10(const unsigned int ADValHR ) {
    unsigned int ad, x1, x2, x3, x4, cd10, b10;
    // 桁ごとの数値を取り出す
    ad = ADValHR;
    x1 = ad / 1000;
    ad = ad - 1000 * x1;
    x2 = ad / 100;
    ad = ad - 100 * x2;
    x3 = ad / 10;
    x4 = ad - 10 * x3;
    // 桁ごとにA-D値を掛け算する
    x1 = ADValHR * x1;
    x2 = (ADValHR * x2 + 5) / 10;		// 小数点以下四捨五入
    x3 = (ADValHR * x3 + 50) / 100;		// 小数点以下四捨五入
    x4 = (ADValHR * x4 + 500) / 1000;	// 小数点以下四捨五入
    x1 = x1 + x2 + x3 + x4;				// 計算結果を合成
    x2 = (x1 + 5) / 10;					// 1/10する. 小数点以下四捨五入
    cd10 = (28 * x2 + 50) / 100;
    x1 = (ADValHR * 4 + 5) / 10;
    x2 = (ADValHR * 5 + 500) / 1000;
    b10 = x1 + x2;
    return -40 + b10 - cd10;			// 湿度値を10倍した整数値
}

void setup() {
	Serial.begin(9600);	
	Serial.println("hello!");
	while( SD.begin() == false ){
		Serial.println("SD ERROR");
		delay(3000);
	}
}

void loop() {
	float hum,hum2,temp,temp2;
	unsigned int wdat;
	byte bdat_H, bdat_L, bdat_CRC;
	File file;
	char filename[] = "sht_hum.csv";
	char Str[16];
	
	SHTInit();

	SHTTSSeq();							// TSシーケンス発行
	SHTWrite(0b00000011);				// measure temprature コマンド送信
	while( digitalRead(portDATA) );		// A-D変換完了を待つ
	bdat_H = SHTRead(0);				// read high byte(ACK)
	bdat_L = SHTRead(0);				// read low byte(ACK)
	bdat_CRC = SHTRead(1);				// read CRC(NOACK)

	wdat = ((unsigned int)bdat_H << 8) | bdat_L;                 // 16ビットに合成
	temp = CalcTMP(wdat);
	temp2= CalcTMP10(wdat)/10.;
	
	SHTTSSeq();							// TSシーケンス発行
	SHTWrite(0b00000101);				// measure humidity コマンド送信
	while( digitalRead(portDATA) );		// A-D変換完了を待つ
	bdat_H = SHTRead(0);				// read high byte(ACK)
	bdat_L = SHTRead(0);				// read low byte(ACK)
	bdat_CRC = SHTRead(1);				// read CRC(NOACK)

	wdat = ((unsigned int)bdat_H << 8) | bdat_L;                 // 16ビットに合成
	hum  = CalcHR(wdat,temp);
	hum2 = CalcHR10(wdat)/10.;
	
	Serial.print("SHT TEMP = ");
	Serial.println( temp , 2 );
	Serial.print("SHT TEMP2= ");
	Serial.println( temp2 , 2 );
	Serial.print("SHT HUM  = ");
	Serial.print( hum , 2 );
	Serial.print(" (0x");
	Serial.print( wdat , HEX);
	Serial.print(")\n");
	Serial.print("SHT HUM2 = ");
	Serial.println( hum2 , 2 );

	Serial.print("writing SD ... ");
	if( (file = SD.open( filename, FILE_WRITE )) != false ){
		file.print( temp , 2 );
		file.print( ", " );
		file.println( hum , 2 );
		file.close();
		Serial.println( "done" );
	} else Serial.println("failed file open");
	delay(3000);
}

